"""
Embedded Python Blocks:

Each time this file is saved, GRC will instantiate the first class it finds
to get ports and parameters of your block. The arguments to __init__  will
be the parameters. All of them are required to have default values!
"""

import numpy as np, scipy as sc
from gnuradio import gr


class sinad(gr.decim_block):
    """
    docstring for block sinad
    """

    def __init__(self, SinadFreq=1750, Fs=48000, fft_size=4096, hop_size=256, refWidth=20):
        gr.decim_block.__init__(
        self,
        name="SINAD",
        in_sig=[np.float32],
        out_sig=[np.float32],
        decim = (hop_size)
        )
        self.Fs = Fs
        self.fRef = SinadFreq
        self.refWidth = refWidth
        self.fft_size = fft_size
        self.hop_size = hop_size
        
        # Initialiser un tampon circulaire pour la fenêtre FFT glissante
        self.buffer = np.zeros(self.fft_size, dtype=np.float32)
        
        # Garantir que work est appelé pour chaque (hop_size) échantillons
        #self.set_output_multiple(hop_size)
 
    def work(self, input_items, output_items):
        in0 = input_items[0]
        out0 = output_items[0]

        if len(in0) < self.hop_size:
        # Ne rien faire si pas assez d'échantillons
            return 0

        # Décaler le tampon circulaire d'un nombre d'échantillons égal à hop_size
        self.buffer = np.roll(self.buffer, -self.hop_size)

        # Ajouter les nouveaux échantillons à la fin du tampon
        self.buffer[-self.hop_size:] = in0[:self.hop_size]
        
        
        # Calculer le SINAD 
        out0[:] = self.__calc_sinad(self.buffer) 
                
        # -------------------------------------------------------------------------------------        
        # code alternatif avec mise à jour du tampon échantillon par échantillon
        
        # Remplir le tampon circulaire avec les échantillons d'entrée
        #for i in range(len(in0)):
        #    self.buffer = np.roll(self.buffer, -1)  # Décaler les anciens échantillons
        #    self.buffer[-1] = in0[i]  # Ajouter le nouvel échantillon à la fin

        # Lorsque le tampon est rempli, calculer le SINAD
        #if i % self.hop_size == 0:
        #    out0[i // self.hop_size] = self.__calc_sinad(self.buffer)
        # -------------------------------------------------------------------------------------
      
        # retourner le nombre de valeurs en sortie, une seule valeur SINAD.
        # (indispensable pour la synchronisation des données)
        return 1

    def __calc_sinad(self, data):
        """ Calcul du SINAD en dB
        """
        # Filtrer les données avec une fenêtre de Kaiser
        window = np.kaiser(np.size(data),8)
        data = data * window
        
        # Calculer la FFT
        psd = np.fft.fft(data)

        # Determiner les index des bins
        n = len(psd)
        # Le signal se situe entre bin1 et bin2
        bin1 = int((self.fRef - self.refWidth / 2) / self.Fs * n)
        bin2 = int((self.fRef + self.refWidth / 2) / self.Fs * n)

        # limitation de la bande passante : 300 Hz - 3000Hz
        bin300hz = int(300 / self.Fs * n)
        bin3000hz = int(3000 / self.Fs * n)

        # calcul des FFT partielles pour le signal et pour le bruit
        signal = psd[bin300hz:bin3000hz]
        noise = np.concatenate((psd[bin300hz:bin1], psd[(bin2 + 1):bin3000hz]))

        # calcul des puissances de signal et de bruit
        Ps = np.sum(np.abs(signal) ** 2)
        Pn = np.sum(np.abs(noise) ** 2)
        
        # Calcul du SINAD : SINAD = 10*log10((Ps + Pn) / Pn)
        sinad = max(10 * np.log10(Ps / Pn), 0) if Ps > 0 else 0
        return sinad